import json
from base_entity import BaseEntity
from base_entity import Property
from base_entity import EntityDict
from request import Request
import trade
import position
import order
import transaction
import primitives



class Account(BaseEntity):
    _summary_format = "Account {id}"
    _name_format = ""

    _properties = [
        Property(
            "id",
            "Account ID",
            "The Account's identifier",
            "primitive",
            "account.AccountID",
            None
        ),
        Property(
            "alias",
            "Alias",
            "Client-assigned alias for the Account. Only provided if the Account has an alias set",
            "primitive",
            "string",
            None
        ),
        Property(
            "currency",
            "Home Currency",
            "The home currency of the Account",
            "primitive",
            "primitives.Currency",
            None
        ),
        Property(
            "balance",
            "Balance",
            "The current balance of the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "createdByUserID",
            "Created by User ID",
            "ID of the user that created the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "createdTime",
            "Create Time",
            "The date/time when the Account was created.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "pl",
            "Profit/Loss",
            "The total profit/loss realized over the lifetime of the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "resettabledPL",
            "Resettable Profit/Loss",
            "The total realized profit/loss for the Account since it was last reset by the client. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "resettabledPLTime",
            "Profit/Loss Reset Time",
            "The date/time that the Account's resettablePL was last reset.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "marginRate",
            "Margin Rate",
            "Client-provided margin rate override for the Account. The effective margin rate of the Account is the lesser of this value and the OANDA margin rate for the Account's division. This value is only provided if a margin rate override exists for the Account.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "marginCallEnterTime",
            "Margin Call Enter Time",
            "The date/time when the Account entered a margin call state. Only provided if the Account is in a margin call.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "marginCallExtensionCount",
            "Margin Call Extension Count",
            "The number of times that the Account's current margin call was extended.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "lastMarginCallExtensionTime",
            "Last Margin Call Extension Time",
            "The date/time of the Account's last margin call extension.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "openTradeCount",
            "Open Trade Count",
            "The number of Trades currently open in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "openPositionCount",
            "Open Position Count",
            "The number of Positions currently open in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "pendingOrderCount",
            "Pending Order Count",
            "The number of Orders currently pending in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "hedgingEnabled",
            "Hedging Enabled",
            "Flag indicating that the Account has hedging enabled.",
            "primitive",
            "boolean",
            None
        ),
        Property(
            "unrealizedPL",
            "Unrealized Profit/Loss",
            "The total unrealized profit/loss for all Trades currently open in the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "NAV",
            "Net Asset Value",
            "The net asset value of the Account. Equal to Account balance + unrealizedPL. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginUsed",
            "Margin Used",
            "Margin currently used for the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginAvailable",
            "Margin Available",
            "Margin available for Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "positionValue",
            "Position Value",
            "The value of the Account's open positions represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutUnrealizedPL",
            "Closeout UPL",
            "The Account's margin closeout unrealized PL.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutNAV",
            "Closeout NAV",
            "The Account's margin closeout NAV.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutMarginUsed",
            "Closeout Margin Used",
            "The Account's margin closeout margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutPercent",
            "Margin Closeout Percentage",
            "The Account's margin closeout percentage. When this value is 1.0 or above the Account is in a margin closeout situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "withdrawalLimit",
            "Withdrawal Limit",
            "The current WithdrawalLimit for the account which will be zero or a positive value indicating how much can be withdrawn from the account.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallMarginUsed",
            "Margin Call Margin Used",
            "The Account's margin call margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallPercent",
            "Margin Call Percentage",
            "The Account's margin call percentage. When this value is 1.0 or above the Account is in a margin call situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "lastTransactionID",
            "Last Transaction ID",
            "The ID of the last Transaction created for the Account.",
            "primitive",
            "transaction.TransactionID",
            None
        ),
        Property(
            "trades",
            "Open Trades",
            "The details of the Trades currently open in the Account.",
            "array_object",
            "TradeSummary",
            None
        ),
        Property(
            "positions",
            "Open Positions",
            "The details of all Positions currently open in the Account.",
            "array_object",
            "Position",
            None
        ),
        Property(
            "orders",
            "Pending Orders",
            "The details of the Orders currently pending in the Account.",
            "array_object",
            "Order",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(Account, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('alias') is not None:
            body['alias'] = \
                data.get('alias')

        if data.get('currency') is not None:
            body['currency'] = \
                data.get('currency')

        if data.get('balance') is not None:
            body['balance'] = \
                data.get('balance')

        if data.get('createdByUserID') is not None:
            body['createdByUserID'] = \
                data.get('createdByUserID')

        if data.get('createdTime') is not None:
            body['createdTime'] = \
                data.get('createdTime')

        if data.get('pl') is not None:
            body['pl'] = \
                data.get('pl')

        if data.get('resettabledPL') is not None:
            body['resettabledPL'] = \
                data.get('resettabledPL')

        if data.get('resettabledPLTime') is not None:
            body['resettabledPLTime'] = \
                data.get('resettabledPLTime')

        if data.get('marginRate') is not None:
            body['marginRate'] = \
                data.get('marginRate')

        if data.get('marginCallEnterTime') is not None:
            body['marginCallEnterTime'] = \
                data.get('marginCallEnterTime')

        if data.get('marginCallExtensionCount') is not None:
            body['marginCallExtensionCount'] = \
                data.get('marginCallExtensionCount')

        if data.get('lastMarginCallExtensionTime') is not None:
            body['lastMarginCallExtensionTime'] = \
                data.get('lastMarginCallExtensionTime')

        if data.get('openTradeCount') is not None:
            body['openTradeCount'] = \
                data.get('openTradeCount')

        if data.get('openPositionCount') is not None:
            body['openPositionCount'] = \
                data.get('openPositionCount')

        if data.get('pendingOrderCount') is not None:
            body['pendingOrderCount'] = \
                data.get('pendingOrderCount')

        if data.get('hedgingEnabled') is not None:
            body['hedgingEnabled'] = \
                data.get('hedgingEnabled')

        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        if data.get('NAV') is not None:
            body['NAV'] = \
                data.get('NAV')

        if data.get('marginUsed') is not None:
            body['marginUsed'] = \
                data.get('marginUsed')

        if data.get('marginAvailable') is not None:
            body['marginAvailable'] = \
                data.get('marginAvailable')

        if data.get('positionValue') is not None:
            body['positionValue'] = \
                data.get('positionValue')

        if data.get('marginCloseoutUnrealizedPL') is not None:
            body['marginCloseoutUnrealizedPL'] = \
                data.get('marginCloseoutUnrealizedPL')

        if data.get('marginCloseoutNAV') is not None:
            body['marginCloseoutNAV'] = \
                data.get('marginCloseoutNAV')

        if data.get('marginCloseoutMarginUsed') is not None:
            body['marginCloseoutMarginUsed'] = \
                data.get('marginCloseoutMarginUsed')

        if data.get('marginCloseoutPercent') is not None:
            body['marginCloseoutPercent'] = \
                data.get('marginCloseoutPercent')

        if data.get('withdrawalLimit') is not None:
            body['withdrawalLimit'] = \
                data.get('withdrawalLimit')

        if data.get('marginCallMarginUsed') is not None:
            body['marginCallMarginUsed'] = \
                data.get('marginCallMarginUsed')

        if data.get('marginCallPercent') is not None:
            body['marginCallPercent'] = \
                data.get('marginCallPercent')

        if data.get('lastTransactionID') is not None:
            body['lastTransactionID'] = \
                data.get('lastTransactionID')

        if data.get('trades') is not None:
            body['trades'] = [
                trade.TradeSummary.from_dict(d)
                for d in data.get('trades')\
            ]

        if data.get('positions') is not None:
            body['positions'] = [
                position.Position.from_dict(d)
                for d in data.get('positions')\
            ]

        if data.get('orders') is not None:
            body['orders'] = [
                order.Order.from_dict(d)
                for d in data.get('orders')\
            ]

        self = Account(**body)

        return self


class AccountState(BaseEntity):
    _summary_format = ""
    _name_format = ""

    _properties = [
        Property(
            "unrealizedPL",
            "Unrealized Profit/Loss",
            "The total unrealized profit/loss for all Trades currently open in the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "NAV",
            "Net Asset Value",
            "The net asset value of the Account. Equal to Account balance + unrealizedPL. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginUsed",
            "Margin Used",
            "Margin currently used for the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginAvailable",
            "Margin Available",
            "Margin available for Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "positionValue",
            "Position Value",
            "The value of the Account's open positions represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutUnrealizedPL",
            "Closeout UPL",
            "The Account's margin closeout unrealized PL.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutNAV",
            "Closeout NAV",
            "The Account's margin closeout NAV.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutMarginUsed",
            "Closeout Margin Used",
            "The Account's margin closeout margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutPercent",
            "Margin Closeout Percentage",
            "The Account's margin closeout percentage. When this value is 1.0 or above the Account is in a margin closeout situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "withdrawalLimit",
            "Withdrawal Limit",
            "The current WithdrawalLimit for the account which will be zero or a positive value indicating how much can be withdrawn from the account.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallMarginUsed",
            "Margin Call Margin Used",
            "The Account's margin call margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallPercent",
            "Margin Call Percentage",
            "The Account's margin call percentage. When this value is 1.0 or above the Account is in a margin call situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "orders",
            "Order States",
            "The price-dependent state of each pending Order in the Account.",
            "array_object",
            "DynamicOrderState",
            None
        ),
        Property(
            "trades",
            "Trade States",
            "The price-dependent state for each open Trade in the Account.",
            "array_object",
            "CalculatedTradeState",
            None
        ),
        Property(
            "positions",
            "Position States",
            "The price-dependent state for each open Position in the Account.",
            "array_object",
            "CalculatedPositionState",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(AccountState, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        if data.get('NAV') is not None:
            body['NAV'] = \
                data.get('NAV')

        if data.get('marginUsed') is not None:
            body['marginUsed'] = \
                data.get('marginUsed')

        if data.get('marginAvailable') is not None:
            body['marginAvailable'] = \
                data.get('marginAvailable')

        if data.get('positionValue') is not None:
            body['positionValue'] = \
                data.get('positionValue')

        if data.get('marginCloseoutUnrealizedPL') is not None:
            body['marginCloseoutUnrealizedPL'] = \
                data.get('marginCloseoutUnrealizedPL')

        if data.get('marginCloseoutNAV') is not None:
            body['marginCloseoutNAV'] = \
                data.get('marginCloseoutNAV')

        if data.get('marginCloseoutMarginUsed') is not None:
            body['marginCloseoutMarginUsed'] = \
                data.get('marginCloseoutMarginUsed')

        if data.get('marginCloseoutPercent') is not None:
            body['marginCloseoutPercent'] = \
                data.get('marginCloseoutPercent')

        if data.get('withdrawalLimit') is not None:
            body['withdrawalLimit'] = \
                data.get('withdrawalLimit')

        if data.get('marginCallMarginUsed') is not None:
            body['marginCallMarginUsed'] = \
                data.get('marginCallMarginUsed')

        if data.get('marginCallPercent') is not None:
            body['marginCallPercent'] = \
                data.get('marginCallPercent')

        if data.get('orders') is not None:
            body['orders'] = [
                order.DynamicOrderState.from_dict(d)
                for d in data.get('orders')\
            ]

        if data.get('trades') is not None:
            body['trades'] = [
                trade.CalculatedTradeState.from_dict(d)
                for d in data.get('trades')\
            ]

        if data.get('positions') is not None:
            body['positions'] = [
                position.CalculatedPositionState.from_dict(d)
                for d in data.get('positions')\
            ]

        self = AccountState(**body)

        return self


class AccountProperties(BaseEntity):
    _summary_format = ""
    _name_format = ""

    _properties = [
        Property(
            "id",
            "ID",
            "The Account's identifier",
            "primitive",
            "account.AccountID",
            None
        ),
        Property(
            "mt4AccountID",
            "MT4 Account ID",
            "The Account's associated MT4 Account ID. This field will not be present if the Account is not an MT4 account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "tags",
            "Account Tags",
            "The Account's tags",
            "array_primitive",
            "string",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(AccountProperties, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('mt4AccountID') is not None:
            body['mt4AccountID'] = \
                data.get('mt4AccountID')

        if data.get('tags') is not None:
            body['tags'] = \
                data.get('tags')

        self = AccountProperties(**body)

        return self


class AccountSummary(BaseEntity):
    _summary_format = ""
    _name_format = ""

    _properties = [
        Property(
            "id",
            "Account ID",
            "The Account's identifier",
            "primitive",
            "account.AccountID",
            None
        ),
        Property(
            "alias",
            "Alias",
            "Client-assigned alias for the Account. Only provided if the Account has an alias set",
            "primitive",
            "string",
            None
        ),
        Property(
            "currency",
            "Home Currency",
            "The home currency of the Account",
            "primitive",
            "primitives.Currency",
            None
        ),
        Property(
            "balance",
            "Balance",
            "The current balance of the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "createdByUserID",
            "Created by User ID",
            "ID of the user that created the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "createdTime",
            "Create Time",
            "The date/time when the Account was created.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "pl",
            "Profit/Loss",
            "The total profit/loss realized over the lifetime of the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "resettabledPL",
            "Resettable Profit/Loss",
            "The total realized profit/loss for the Account since it was last reset by the client. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "resettabledPLTime",
            "Profit/Loss Reset Time",
            "The date/time that the Account's resettablePL was last reset.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "marginRate",
            "Margin Rate",
            "Client-provided margin rate override for the Account. The effective margin rate of the Account is the lesser of this value and the OANDA margin rate for the Account's division. This value is only provided if a margin rate override exists for the Account.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "marginCallEnterTime",
            "Margin Call Enter Time",
            "The date/time when the Account entered a margin call state. Only provided if the Account is in a margin call.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "marginCallExtensionCount",
            "Margin Call Extension Count",
            "The number of times that the Account's current margin call was extended.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "lastMarginCallExtensionTime",
            "Last Margin Call Extension Time",
            "The date/time of the Account's last margin call extension.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "openTradeCount",
            "Open Trade Count",
            "The number of Trades currently open in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "openPositionCount",
            "Open Position Count",
            "The number of Positions currently open in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "pendingOrderCount",
            "Pending Order Count",
            "The number of Orders currently pending in the Account.",
            "primitive",
            "integer",
            None
        ),
        Property(
            "hedgingEnabled",
            "Hedging Enabled",
            "Flag indicating that the Account has hedging enabled.",
            "primitive",
            "boolean",
            None
        ),
        Property(
            "unrealizedPL",
            "Unrealized Profit/Loss",
            "The total unrealized profit/loss for all Trades currently open in the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "NAV",
            "Net Asset Value",
            "The net asset value of the Account. Equal to Account balance + unrealizedPL. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginUsed",
            "Margin Used",
            "Margin currently used for the Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginAvailable",
            "Margin Available",
            "Margin available for Account. Represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "positionValue",
            "Position Value",
            "The value of the Account's open positions represented in the Account's home currency.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutUnrealizedPL",
            "Closeout UPL",
            "The Account's margin closeout unrealized PL.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutNAV",
            "Closeout NAV",
            "The Account's margin closeout NAV.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutMarginUsed",
            "Closeout Margin Used",
            "The Account's margin closeout margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCloseoutPercent",
            "Margin Closeout Percentage",
            "The Account's margin closeout percentage. When this value is 1.0 or above the Account is in a margin closeout situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "withdrawalLimit",
            "Withdrawal Limit",
            "The current WithdrawalLimit for the account which will be zero or a positive value indicating how much can be withdrawn from the account.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallMarginUsed",
            "Margin Call Margin Used",
            "The Account's margin call margin used.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "marginCallPercent",
            "Margin Call Percentage",
            "The Account's margin call percentage. When this value is 1.0 or above the Account is in a margin call situation.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "lastTransactionID",
            "Last Transaction ID",
            "The ID of the last Transaction created for the Account.",
            "primitive",
            "transaction.TransactionID",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(AccountSummary, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('alias') is not None:
            body['alias'] = \
                data.get('alias')

        if data.get('currency') is not None:
            body['currency'] = \
                data.get('currency')

        if data.get('balance') is not None:
            body['balance'] = \
                data.get('balance')

        if data.get('createdByUserID') is not None:
            body['createdByUserID'] = \
                data.get('createdByUserID')

        if data.get('createdTime') is not None:
            body['createdTime'] = \
                data.get('createdTime')

        if data.get('pl') is not None:
            body['pl'] = \
                data.get('pl')

        if data.get('resettabledPL') is not None:
            body['resettabledPL'] = \
                data.get('resettabledPL')

        if data.get('resettabledPLTime') is not None:
            body['resettabledPLTime'] = \
                data.get('resettabledPLTime')

        if data.get('marginRate') is not None:
            body['marginRate'] = \
                data.get('marginRate')

        if data.get('marginCallEnterTime') is not None:
            body['marginCallEnterTime'] = \
                data.get('marginCallEnterTime')

        if data.get('marginCallExtensionCount') is not None:
            body['marginCallExtensionCount'] = \
                data.get('marginCallExtensionCount')

        if data.get('lastMarginCallExtensionTime') is not None:
            body['lastMarginCallExtensionTime'] = \
                data.get('lastMarginCallExtensionTime')

        if data.get('openTradeCount') is not None:
            body['openTradeCount'] = \
                data.get('openTradeCount')

        if data.get('openPositionCount') is not None:
            body['openPositionCount'] = \
                data.get('openPositionCount')

        if data.get('pendingOrderCount') is not None:
            body['pendingOrderCount'] = \
                data.get('pendingOrderCount')

        if data.get('hedgingEnabled') is not None:
            body['hedgingEnabled'] = \
                data.get('hedgingEnabled')

        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        if data.get('NAV') is not None:
            body['NAV'] = \
                data.get('NAV')

        if data.get('marginUsed') is not None:
            body['marginUsed'] = \
                data.get('marginUsed')

        if data.get('marginAvailable') is not None:
            body['marginAvailable'] = \
                data.get('marginAvailable')

        if data.get('positionValue') is not None:
            body['positionValue'] = \
                data.get('positionValue')

        if data.get('marginCloseoutUnrealizedPL') is not None:
            body['marginCloseoutUnrealizedPL'] = \
                data.get('marginCloseoutUnrealizedPL')

        if data.get('marginCloseoutNAV') is not None:
            body['marginCloseoutNAV'] = \
                data.get('marginCloseoutNAV')

        if data.get('marginCloseoutMarginUsed') is not None:
            body['marginCloseoutMarginUsed'] = \
                data.get('marginCloseoutMarginUsed')

        if data.get('marginCloseoutPercent') is not None:
            body['marginCloseoutPercent'] = \
                data.get('marginCloseoutPercent')

        if data.get('withdrawalLimit') is not None:
            body['withdrawalLimit'] = \
                data.get('withdrawalLimit')

        if data.get('marginCallMarginUsed') is not None:
            body['marginCallMarginUsed'] = \
                data.get('marginCallMarginUsed')

        if data.get('marginCallPercent') is not None:
            body['marginCallPercent'] = \
                data.get('marginCallPercent')

        if data.get('lastTransactionID') is not None:
            body['lastTransactionID'] = \
                data.get('lastTransactionID')

        self = AccountSummary(**body)

        return self


class AccountChanges(BaseEntity):
    _summary_format = ""
    _name_format = ""

    _properties = [
        Property(
            "ordersCreated",
            "Orders Created",
            "The Orders created. These Orders may have been filled, cancelled or triggered in the same period.",
            "array_object",
            "Order",
            None
        ),
        Property(
            "ordersCancelled",
            "Orders Cancelled",
            "The Orders cancelled.",
            "array_object",
            "Order",
            None
        ),
        Property(
            "ordersFilled",
            "Orders Filled",
            "The Orders filled.",
            "array_object",
            "Order",
            None
        ),
        Property(
            "ordersTriggered",
            "Orders Triggered",
            "The Orders triggered.",
            "array_object",
            "Order",
            None
        ),
        Property(
            "tradesOpened",
            "Trades Opened",
            "The Trades opened.",
            "array_object",
            "Trade",
            None
        ),
        Property(
            "tradesReduced",
            "Trades Reduced",
            "The Trades reduced.",
            "array_object",
            "Trade",
            None
        ),
        Property(
            "tradesClosed",
            "Trades Closed",
            "The Trades closed.",
            "array_object",
            "Trade",
            None
        ),
        Property(
            "positions",
            "Positions",
            "The Positions changed.",
            "array_object",
            "Position",
            None
        ),
        Property(
            "transactions",
            "Transactions",
            "The Transactions that have been generated.",
            "array_object",
            "Transaction",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(AccountChanges, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('ordersCreated') is not None:
            body['ordersCreated'] = [
                order.Order.from_dict(d)
                for d in data.get('ordersCreated')\
            ]

        if data.get('ordersCancelled') is not None:
            body['ordersCancelled'] = [
                order.Order.from_dict(d)
                for d in data.get('ordersCancelled')\
            ]

        if data.get('ordersFilled') is not None:
            body['ordersFilled'] = [
                order.Order.from_dict(d)
                for d in data.get('ordersFilled')\
            ]

        if data.get('ordersTriggered') is not None:
            body['ordersTriggered'] = [
                order.Order.from_dict(d)
                for d in data.get('ordersTriggered')\
            ]

        if data.get('tradesOpened') is not None:
            body['tradesOpened'] = [
                trade.Trade.from_dict(d)
                for d in data.get('tradesOpened')\
            ]

        if data.get('tradesReduced') is not None:
            body['tradesReduced'] = [
                trade.Trade.from_dict(d)
                for d in data.get('tradesReduced')\
            ]

        if data.get('tradesClosed') is not None:
            body['tradesClosed'] = [
                trade.Trade.from_dict(d)
                for d in data.get('tradesClosed')\
            ]

        if data.get('positions') is not None:
            body['positions'] = [
                position.Position.from_dict(d)
                for d in data.get('positions')\
            ]

        if data.get('transactions') is not None:
            body['transactions'] = [
                transaction.Transaction.from_dict(d)
                for d in data.get('transactions')\
            ]

        self = AccountChanges(**body)

        return self

class EntitySpec(object):
    Account = Account
    AccountState = AccountState
    AccountProperties = AccountProperties
    AccountSummary = AccountSummary
    AccountChanges = AccountChanges

    def __init__(self, ctx):
        self.ctx = ctx


    def list(
        self,
        **kwargs
    ):
        """List Accounts

        Get a list of all Accounts authorized for the provided token.

        Parameters
        ----------
        """


        request = Request(
            'GET',
            '/v3/accounts'
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('accounts') is not None:
                parsed_body['accounts'] = [
                    AccountProperties.from_dict(d)
                    for d in jbody.get('accounts')
                ]


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def get(
        self,
        accountID,
        **kwargs
    ):
        """Account Details

        Get the full details for a single Account that a client has access to.
        Full pending Order, open Trade and open Position representations are
        provided.

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('account') is not None:
                parsed_body['account'] = \
                    Account.from_dict(
                        jbody['account']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def summary(
        self,
        accountID,
        **kwargs
    ):
        """Account Summary

        Get a summary for a single Account that a client has access to.

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/summary'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('account') is not None:
                parsed_body['account'] = \
                    AccountSummary.from_dict(
                        jbody['account']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def instruments(
        self,
        accountID,
        **kwargs
    ):
        """Account Instruments

        Get the list of tradeable instruments for the given Account.

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch
        instruments : array, optional
            List of instruments to query specifically.
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/instruments'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_param(
            'instruments',
            kwargs.get('instruments')
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('instruments') is not None:
                parsed_body['instruments'] = [
                    primitives.Instrument.from_dict(d)
                    for d in jbody.get('instruments')\
                ]


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def configure(
        self,
        accountID,
        **kwargs
    ):
        """Configure Account

        Set the client-configurable portions of an Account.

        Parameters
        ----------
        accountID : 
            ID of the Account to configure
        alias : string, optional
            Client-defined alias (name) for the Account
        marginRate : None, optional
            The string representation of a decimal number.
        """


        request = Request(
            'PATCH',
            '/v3/accounts/{accountID}/configuration'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        body = EntityDict()

        body.set('alias', kwargs.get('alias'))

        body.set('marginRate', kwargs.get('marginRate'))

        request.set_body_dict(body.dict)

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('configureTransaction') is not None:
                parsed_body['configureTransaction'] = \
                    transaction.ClientConfigureTransaction.from_dict(
                        jbody['configureTransaction']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 400:
            if jbody.get('configureRejectTransaction') is not None:
                parsed_body['configureRejectTransaction'] = \
                    transaction.ClientConfigureRejectTransaction.from_dict(
                        jbody['configureRejectTransaction']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')

            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def changes(
        self,
        accountID,
        **kwargs
    ):
        """Poll Account Updates

        Endpoint used to poll an Account for its current state and changes
        since a specified TransactionID.

        Parameters
        ----------
        accountID : 
            ID of the Account to get changes for
        sinceTransactionID : , optional
            ID of the Transaction to get Account changes since.
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/changes'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_param(
            'sinceTransactionID',
            kwargs.get('sinceTransactionID')
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('changes') is not None:
                parsed_body['changes'] = \
                    AccountChanges.from_dict(
                        jbody['changes']
                    )

            if jbody.get('state') is not None:
                parsed_body['state'] = \
                    AccountState.from_dict(
                        jbody['state']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 416:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response




import json
import re


class Property(object):
    def __init__(
        self,
        name,
        displayName,
        description,
        typeClass,
        typeName,
        default
    ):
        self.name = name
        self.displayName = displayName
        self.description = description
        self.typeClass = typeClass
        self.typeName = typeName
        self.default = default


class Field(Property):
    def __init__(self, property, value):
        super(Field, self).__init__(
            property.name,
            property.displayName, 
            property.description, 
            property.typeClass, 
            property.typeName,
            property.default
        )

        self.value = value

    def __str__(self):
        if self.typeClass.startswith("array"):
            return "{} = [{}]".format(
                self.displayName, len(self.value)
            )

        if self.typeClass.startswith("object"):
            v = str(self.value)
            v = v.strip()
            v = "{ " + ", ".join(v.split("\n")) + " }"

            return "{} = {}".format(
                self.displayName, v
            )

        return "{} = {}".format(
            self.displayName,
            self.value
        )


class EntityDict(object):
    def __init__(self):
        self.dict = {}

    def __len__(self):
        return len(self.dict)

    def prop_dict_value(self, value):
        if hasattr(value, 'dict'):
            return value.dict()
        return value

    def set(self, key, value):
        if value is None:
            return

        if isinstance(value, list):
            self.dict[key] = [self.prop_dict_value(v) for v in value]
        else:
            self.dict[key] = self.prop_dict_value(value)


class BaseEntity(object):
    _properties = []

    def __init__(self):
        pass

    def fields(self):
        for prop in self._properties:
            value = getattr(self, prop.name)
            if value is not None:
                yield Field(prop, value)

    def name(self):
        name_string = self._name_format

        for m in re.findall(r'{[^}]+}', name_string):
            key = m[1:-1]
            value = getattr(self, key)
            if value is not None:
                name_string = name_string.replace(m, value)

        return name_string

    def summary(self):
        summary_string = self._summary_format

        for m in re.findall(r'{[^}]+}', summary_string):
            key = m[1:-1]
            value = getattr(self, key)
            if value is not None:
                summary_string = summary_string.replace(m, value)

        return summary_string

    def title(self):
        name = self.name()
        summary  = self.summary()
        
        title_string = name

        if len(name) > 0 and len(summary) > 0:
            title_string += ": "

        title_string += summary

        return title_string

    def __str__(self):
        s = self.title()

        if len(s) > 0:
            s += "\n"

        for field in self.fields():
            s += str(field) + "\n"

        return s

    def dict(self):
        entity = EntityDict()
        for field in self.fields():
            entity.set(field.name, field.value)
        return entity.dict

    def json(self):
        return json.dumps(self.dict())

    def diff(self, other):
        for property in self._properties:
            self_value = getattr(self, property.name)
            other_value = getattr(other, property.name)

            # print property.name, property.typeClass, self_value, other_value

            if self_value is None:
                continue

            assert \
                other_value is not None, \
                "Did not expected {}.{} to be unset".format(
                    self.__class__.__name__,
                    property.name
                )

            if property.typeClass == "primitive":
                assert \
                    self_value == other_value, \
                    "Expected value for {}.{} is {}, got {}".format(
                        self.__class__.__name__,
                        property.name,
                        self_value,
                        other_value
                    )
            elif property.typeClass == "object":
                self_value.diff(other_value)

        return True



import json
from base_entity import BaseEntity
from base_entity import Property
from base_entity import EntityDict
from request import Request
import transaction
import order



class Trade(BaseEntity):
    _summary_format = "{initialUnits} of {instrument} @ {price}"
    _name_format = "Trade {id}"

    _properties = [
        Property(
            "id",
            "Trade ID",
            "The Trade's identifier, unique within the Trade's Account.",
            "primitive",
            "trade.TradeID",
            None
        ),
        Property(
            "instrument",
            "Instrument",
            "The Trade's Instrument.",
            "primitive",
            "primitives.InstrumentName",
            None
        ),
        Property(
            "price",
            "Fill Price",
            "The execution price of the Trade.",
            "primitive",
            "pricing.PriceValue",
            None
        ),
        Property(
            "openTime",
            "Open Time",
            "The date/time when the Trade was opened.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "state",
            "State",
            "The current state of the Trade.",
            "primitive",
            "trade.TradeState",
            None
        ),
        Property(
            "initialUnits",
            "Initial Trade Units",
            "The initial size of the Trade. Negative values indicate a short Trade, and positive values indicate a long Trade.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "currentUnits",
            "Current Open Trade Units",
            "The number of units currently open for the Trade. This value is reduced to 0.0 as the Trade is closed.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "realizedPL",
            "Realized Profit/Loss",
            "The total profit/loss realized on the closed portion of the Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "unrealizedPL",
            "Unrealized Profit/Loss",
            "The unrealized profit/loss on the open portion of the Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "closingTransactionIDs",
            "Closing Transaction IDs",
            "The IDs of the Transactions that have closed portions of this Trade.",
            "array_primitive",
            "TransactionID",
            None
        ),
        Property(
            "financing",
            "Financing",
            "The financing paid/collected for this Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "closeTime",
            "Close Time",
            "The date/time when the Trade was fully closed. Only provided for Trades whose state is CLOSED.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "clientExtensions",
            "Client Extensions",
            "The client extensions of the Trade.",
            "object",
            "transaction.ClientExtensions",
            None
        ),
        Property(
            "takeProfitOrder",
            "Take Profit Order",
            "Full representation of the Trade's Take Profit Order, only provided if such an Order exists.",
            "object",
            "order.TakeProfitOrder",
            None
        ),
        Property(
            "stopLossOrder",
            "Stop Loss Order",
            "Full representation of the Trade's Stop Loss Order, only provided if such an Order exists.",
            "object",
            "order.StopLossOrder",
            None
        ),
        Property(
            "trailingStopLossOrder",
            "Trailing Stop Loss Order",
            "Full representation of the Trade's Trailing Stop Loss Order, only provided if such an Order exists.",
            "object",
            "order.TrailingStopLossOrder",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(Trade, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('instrument') is not None:
            body['instrument'] = \
                data.get('instrument')

        if data.get('price') is not None:
            body['price'] = \
                data.get('price')

        if data.get('openTime') is not None:
            body['openTime'] = \
                data.get('openTime')

        if data.get('state') is not None:
            body['state'] = \
                data.get('state')

        if data.get('initialUnits') is not None:
            body['initialUnits'] = \
                data.get('initialUnits')

        if data.get('currentUnits') is not None:
            body['currentUnits'] = \
                data.get('currentUnits')

        if data.get('realizedPL') is not None:
            body['realizedPL'] = \
                data.get('realizedPL')

        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        if data.get('closingTransactionIDs') is not None:
            body['closingTransactionIDs'] = \
                data.get('closingTransactionIDs')

        if data.get('financing') is not None:
            body['financing'] = \
                data.get('financing')

        if data.get('closeTime') is not None:
            body['closeTime'] = \
                data.get('closeTime')

        if data.get('clientExtensions') is not None:
            body['clientExtensions'] = \
                transaction.ClientExtensions.from_dict(
                    data['clientExtensions']
                )

        if data.get('takeProfitOrder') is not None:
            body['takeProfitOrder'] = \
                order.TakeProfitOrder.from_dict(
                    data['takeProfitOrder']
                )

        if data.get('stopLossOrder') is not None:
            body['stopLossOrder'] = \
                order.StopLossOrder.from_dict(
                    data['stopLossOrder']
                )

        if data.get('trailingStopLossOrder') is not None:
            body['trailingStopLossOrder'] = \
                order.TrailingStopLossOrder.from_dict(
                    data['trailingStopLossOrder']
                )

        self = Trade(**body)

        return self


class TradeSummary(BaseEntity):
    _summary_format = "{initialUnits} of {instrument} @ {price}"
    _name_format = "Trade {id}"

    _properties = [
        Property(
            "id",
            "Trade ID",
            "The Trade's identifier, unique within the Trade's Account.",
            "primitive",
            "trade.TradeID",
            None
        ),
        Property(
            "instrument",
            "Instrument",
            "The Trade's Instrument.",
            "primitive",
            "primitives.InstrumentName",
            None
        ),
        Property(
            "price",
            "Fill Price",
            "The execution price of the Trade.",
            "primitive",
            "pricing.PriceValue",
            None
        ),
        Property(
            "openTime",
            "Open Time",
            "The date/time when the Trade was opened.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "state",
            "State",
            "The current state of the Trade.",
            "primitive",
            "trade.TradeState",
            None
        ),
        Property(
            "initialUnits",
            "Initial Trade Units",
            "The initial size of the Trade. Negative values indicate a short Trade, and positive values indicate a long Trade.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "currentUnits",
            "Current Open Trade Units",
            "The number of units currently open for the Trade. This value is reduced to 0.0 as the Trade is closed.",
            "primitive",
            "primitives.DecimalNumber",
            None
        ),
        Property(
            "realizedPL",
            "Realized Profit/Loss",
            "The total profit/loss realized on the closed portion of the Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "unrealizedPL",
            "Unrealized Profit/Loss",
            "The unrealized profit/loss on the open portion of the Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "closingTransactionIDs",
            "Closing Transaction IDs",
            "The IDs of the Transactions that have closed portions of this Trade.",
            "array_primitive",
            "TransactionID",
            None
        ),
        Property(
            "financing",
            "Financing",
            "The financing paid/collected for this Trade.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
        Property(
            "closeTime",
            "Close Time",
            "The date/time when the Trade was fully closed. Only provided for Trades whose state is CLOSED.",
            "primitive",
            "primitives.DateTime",
            None
        ),
        Property(
            "clientExtensions",
            "Client Extensions",
            "The client extensions of the Trade.",
            "object",
            "transaction.ClientExtensions",
            None
        ),
        Property(
            "takeProfitOrderID",
            "Take Profit Order ID",
            "ID of the Trade's Take Profit Order, only provided if such an Order exists.",
            "primitive",
            "order.OrderID",
            None
        ),
        Property(
            "stopLossOrderID",
            "Stop Loss Order ID",
            "ID of the Trade's Stop Loss Order, only provided if such an Order exists.",
            "primitive",
            "order.OrderID",
            None
        ),
        Property(
            "trailingStopLossOrderID",
            "Trailing Stop Loss Order ID",
            "ID of the Trade's Trailing Stop Loss Order, only provided if such an Order exists.",
            "primitive",
            "order.OrderID",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(TradeSummary, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('instrument') is not None:
            body['instrument'] = \
                data.get('instrument')

        if data.get('price') is not None:
            body['price'] = \
                data.get('price')

        if data.get('openTime') is not None:
            body['openTime'] = \
                data.get('openTime')

        if data.get('state') is not None:
            body['state'] = \
                data.get('state')

        if data.get('initialUnits') is not None:
            body['initialUnits'] = \
                data.get('initialUnits')

        if data.get('currentUnits') is not None:
            body['currentUnits'] = \
                data.get('currentUnits')

        if data.get('realizedPL') is not None:
            body['realizedPL'] = \
                data.get('realizedPL')

        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        if data.get('closingTransactionIDs') is not None:
            body['closingTransactionIDs'] = \
                data.get('closingTransactionIDs')

        if data.get('financing') is not None:
            body['financing'] = \
                data.get('financing')

        if data.get('closeTime') is not None:
            body['closeTime'] = \
                data.get('closeTime')

        if data.get('clientExtensions') is not None:
            body['clientExtensions'] = \
                transaction.ClientExtensions.from_dict(
                    data['clientExtensions']
                )

        if data.get('takeProfitOrderID') is not None:
            body['takeProfitOrderID'] = \
                data.get('takeProfitOrderID')

        if data.get('stopLossOrderID') is not None:
            body['stopLossOrderID'] = \
                data.get('stopLossOrderID')

        if data.get('trailingStopLossOrderID') is not None:
            body['trailingStopLossOrderID'] = \
                data.get('trailingStopLossOrderID')

        self = TradeSummary(**body)

        return self


class CalculatedTradeState(BaseEntity):
    _summary_format = ""
    _name_format = ""

    _properties = [
        Property(
            "id",
            "Trade ID",
            "The Trade's ID.",
            "primitive",
            "trade.TradeID",
            None
        ),
        Property(
            "unrealizedPL",
            "Trade UPL",
            "The Trade's unrealized profit/loss.",
            "primitive",
            "primitives.AccountUnits",
            None
        ),
    ]

    def __init__(self, **kwargs):
        super(CalculatedTradeState, self).__init__()
        for prop in self._properties:
            setattr(self, prop.name, kwargs.get(prop.name, prop.default))

    @staticmethod
    def from_dict(data):

        body = {}
        if data.get('id') is not None:
            body['id'] = \
                data.get('id')

        if data.get('unrealizedPL') is not None:
            body['unrealizedPL'] = \
                data.get('unrealizedPL')

        self = CalculatedTradeState(**body)

        return self

class EntitySpec(object):
    Trade = Trade
    TradeSummary = TradeSummary
    CalculatedTradeState = CalculatedTradeState

    def __init__(self, ctx):
        self.ctx = ctx


    def list(
        self,
        accountID,
        **kwargs
    ):
        """List Trades

        Get a list of Trades for an Account

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch Trades for.
        ids : array, optional
            List of Trade IDs to retrieve.
        state : , optional
            The state to filter the requested Trades by.
        instrument : , optional
            The instrument to filter the requested Trades by.
        count : integer, optional
            The maximum number of Trades to return.
        beforeID : , optional
            The maximum Trade ID to return. If not provided the most recent
            Trades in the Account are returned.
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/trades'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_param(
            'ids',
            kwargs.get('ids')
        )

        request.set_param(
            'state',
            kwargs.get('state')
        )

        request.set_param(
            'instrument',
            kwargs.get('instrument')
        )

        request.set_param(
            'count',
            kwargs.get('count')
        )

        request.set_param(
            'beforeID',
            kwargs.get('beforeID')
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('trades') is not None:
                parsed_body['trades'] = [
                    Trade.from_dict(d)
                    for d in jbody.get('trades')
                ]

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def list_open(
        self,
        accountID,
        **kwargs
    ):
        """List Open Trades

        Get the list of open Trades for an Account

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch Trades for.
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/openTrades'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('trades') is not None:
                parsed_body['trades'] = [
                    Trade.from_dict(d)
                    for d in jbody.get('trades')
                ]

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def get(
        self,
        accountID,
        tradeID,
        **kwargs
    ):
        """Trade Details

        Get the details of a specific Trade in an Account

        Parameters
        ----------
        accountID : 
            ID of the Account to fetch Trades for.
        tradeID : 
            ID of the Trade to fetch
        """


        request = Request(
            'GET',
            '/v3/accounts/{accountID}/trades/{tradeID}'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_path_param(
            'tradeID',
            tradeID
        )

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('trade') is not None:
                parsed_body['trade'] = \
                    Trade.from_dict(
                        jbody['trade']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def close(
        self,
        accountID,
        tradeID,
        **kwargs
    ):
        """Close Trade

        Close (partially or fully) a specific open Trade in an Account

        Parameters
        ----------
        accountID : 
            ID of the Account to close a Trade in.
        tradeID : 
            ID of the Trade to close.
        units : string, optional
            Indication of how much of the Trade to close. Either the string
            "ALL" (indicating that all of the Trade should be closed), or a
            DecimalNumber representing the number of units of the open Trade to
            Close using a TradeClose MarketOrder. The units specified must
            always be positive, and the magnitude of the value cannot exceed
            the magnitude of the Trade's open units.
        """


        request = Request(
            'PUT',
            '/v3/accounts/{accountID}/trades/{tradeID}/close'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_path_param(
            'tradeID',
            tradeID
        )

        body = EntityDict()

        body.set('units', kwargs.get('units'))

        request.set_body_dict(body.dict)

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('orderCreateTransaction') is not None:
                parsed_body['orderCreateTransaction'] = \
                    transaction.MarketOrderTransaction.from_dict(
                        jbody['orderCreateTransaction']
                    )

            if jbody.get('orderFillTransaction') is not None:
                parsed_body['orderFillTransaction'] = \
                    transaction.OrderFillTransaction.from_dict(
                        jbody['orderFillTransaction']
                    )

            if jbody.get('orderCancelTransaction') is not None:
                parsed_body['orderCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['orderCancelTransaction']
                    )

            if jbody.get('relatedTransactionIDs') is not None:
                parsed_body['relatedTransactionIDs'] = \
                    jbody.get('relatedTransactionIDs')

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 400:
            if jbody.get('orderRejectTransaction') is not None:
                parsed_body['orderRejectTransaction'] = \
                    transaction.MarketOrderRejectTransaction.from_dict(
                        jbody['orderRejectTransaction']
                    )

            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('orderRejectTransaction') is not None:
                parsed_body['orderRejectTransaction'] = \
                    transaction.MarketOrderRejectTransaction.from_dict(
                        jbody['orderRejectTransaction']
                    )

            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def set_client_extensions(
        self,
        accountID,
        tradeID,
        **kwargs
    ):
        """Set Trade Client Extensions

        Update the Client Extensions for a Trade

        Parameters
        ----------
        accountID : 
            ID of the Account.
        tradeID : 
            ID of the Trade to update the Client Extension of.
        clientExtensions : None, optional
            The Client Extensions to update the Trade with.
        """


        request = Request(
            'PUT',
            '/v3/accounts/{accountID}/trades/{tradeID}/clientExtensions'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_path_param(
            'tradeID',
            tradeID
        )

        body = EntityDict()

        body.set('clientExtensions', kwargs.get('clientExtensions'))

        request.set_body_dict(body.dict)

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('tradeClientExtensionsModifyTransaction') is not None:
                parsed_body['tradeClientExtensionsModifyTransaction'] = \
                    transaction.TradeClientExtensionsModifyTransaction.from_dict(
                        jbody['tradeClientExtensionsModifyTransaction']
                    )

            if jbody.get('relatedTransactionIDs') is not None:
                parsed_body['relatedTransactionIDs'] = \
                    jbody.get('relatedTransactionIDs')

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 400:
            if jbody.get('tradeClientExtensionsModifyRejectTransaction') is not None:
                parsed_body['tradeClientExtensionsModifyRejectTransaction'] = \
                    transaction.TradeClientExtensionsModifyRejectTransaction.from_dict(
                        jbody['tradeClientExtensionsModifyRejectTransaction']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')

            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response


    def set_dependent_orders(
        self,
        accountID,
        tradeID,
        **kwargs
    ):
        """Set Dependent Orders

        Create, replace and cancel a Trade's dependent Orders (Take Profit,
        Stop Loss and Trailing Stop Loss) through the Trade itself

        Parameters
        ----------
        accountID : 
            ID of the Account.
        tradeID : 
            ID of the Trade to modify the dependent Orders of.
        takeProfit : None, optional
            The specification of the Take Profit to create/modify/cancel. If
            takeProfit is set to null, the Take Profit Order will be cancelled
            if it exists. If takeProfit is not provided, the exisiting Take
            Profit Order will not be modified. If a sub-field of takeProfit is
            not specified, that field will be set to a default value on create,
            and be inherited by the replacing order on modify.
        stopLoss : None, optional
            The specification of the Stop Loss to create/modify/cancel. If
            stopLoss is set to null, the Stop Loss Order will be cancelled if
            it exists. If stopLoss is not provided, the exisiting Stop Loss
            Order will not be modified. If a sub-field of stopLoss is not
            specified, that field will be set to a default value on create, and
            be inherited by the replacing order on modify.
        trailingStopLoss : None, optional
            The specification of the Trailing Stop Loss to
            create/modify/cancel. If trailingStopLoss is set to null, the
            Trailing Stop Loss Order will be cancelled if it exists. If
            trailingStopLoss is not provided, the exisiting Trailing Stop Loss
            Order will not be modified. If a sub-field of trailngStopLoss is
            not specified, that field will be set to a default value on create,
            and be inherited by the replacing order on modify.
        """


        request = Request(
            'PUT',
            '/v3/accounts/{accountID}/trades/{tradeID}/orders'
        )

        request.set_path_param(
            'accountID',
            accountID
        )

        request.set_path_param(
            'tradeID',
            tradeID
        )

        body = EntityDict()

        body.set('takeProfit', kwargs.get('takeProfit'))

        body.set('stopLoss', kwargs.get('stopLoss'))

        body.set('trailingStopLoss', kwargs.get('trailingStopLoss'))

        request.set_body_dict(body.dict)

        response = self.ctx.request(request)


        if response.content_type is None:
            return response

        if not response.content_type.startswith("application/json"):
            return response

        jbody = json.loads(response.raw_body)

        parsed_body = {}

        if response.status is 200:
            if jbody.get('takeProfitOrderCancelTransaction') is not None:
                parsed_body['takeProfitOrderCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['takeProfitOrderCancelTransaction']
                    )

            if jbody.get('takeProfitOrderTransaction') is not None:
                parsed_body['takeProfitOrderTransaction'] = \
                    transaction.TakeProfitOrderTransaction.from_dict(
                        jbody['takeProfitOrderTransaction']
                    )

            if jbody.get('takeProfitOrderFillTransaction') is not None:
                parsed_body['takeProfitOrderFillTransaction'] = \
                    transaction.OrderFillTransaction.from_dict(
                        jbody['takeProfitOrderFillTransaction']
                    )

            if jbody.get('takeProfitOrderCreatedCancelTransaction') is not None:
                parsed_body['takeProfitOrderCreatedCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['takeProfitOrderCreatedCancelTransaction']
                    )

            if jbody.get('stopLossOrderCancelTransaction') is not None:
                parsed_body['stopLossOrderCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['stopLossOrderCancelTransaction']
                    )

            if jbody.get('stopLossOrderTransaction') is not None:
                parsed_body['stopLossOrderTransaction'] = \
                    transaction.StopLossOrderTransaction.from_dict(
                        jbody['stopLossOrderTransaction']
                    )

            if jbody.get('stopLossOrderFillTransaction') is not None:
                parsed_body['stopLossOrderFillTransaction'] = \
                    transaction.OrderFillTransaction.from_dict(
                        jbody['stopLossOrderFillTransaction']
                    )

            if jbody.get('stopLossOrderCreatedCancelTransaction') is not None:
                parsed_body['stopLossOrderCreatedCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['stopLossOrderCreatedCancelTransaction']
                    )

            if jbody.get('trailingStopLossOrderCancelTransaction') is not None:
                parsed_body['trailingStopLossOrderCancelTransaction'] = \
                    transaction.OrderCancelTransaction.from_dict(
                        jbody['trailingStopLossOrderCancelTransaction']
                    )

            if jbody.get('trailingStopLossOrderTransaction') is not None:
                parsed_body['trailingStopLossOrderTransaction'] = \
                    transaction.TrailingStopLossOrderTransaction.from_dict(
                        jbody['trailingStopLossOrderTransaction']
                    )

            if jbody.get('relatedTransactionIDs') is not None:
                parsed_body['relatedTransactionIDs'] = \
                    jbody.get('relatedTransactionIDs')

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')


        if response.status is 400:
            if jbody.get('takeProfitOrderCancelRejectTransaction') is not None:
                parsed_body['takeProfitOrderCancelRejectTransaction'] = \
                    transaction.OrderCancelRejectTransaction.from_dict(
                        jbody['takeProfitOrderCancelRejectTransaction']
                    )

            if jbody.get('takeProfitOrderRejectTransaction') is not None:
                parsed_body['takeProfitOrderRejectTransaction'] = \
                    transaction.TakeProfitOrderRejectTransaction.from_dict(
                        jbody['takeProfitOrderRejectTransaction']
                    )

            if jbody.get('stopLossOrderCancelRejectTransaction') is not None:
                parsed_body['stopLossOrderCancelRejectTransaction'] = \
                    transaction.OrderCancelRejectTransaction.from_dict(
                        jbody['stopLossOrderCancelRejectTransaction']
                    )

            if jbody.get('stopLossOrderRejectTransaction') is not None:
                parsed_body['stopLossOrderRejectTransaction'] = \
                    transaction.StopLossOrderRejectTransaction.from_dict(
                        jbody['stopLossOrderRejectTransaction']
                    )

            if jbody.get('trailingStopLossOrderCancelRejectTransaction') is not None:
                parsed_body['trailingStopLossOrderCancelRejectTransaction'] = \
                    transaction.OrderCancelRejectTransaction.from_dict(
                        jbody['trailingStopLossOrderCancelRejectTransaction']
                    )

            if jbody.get('trailingStopLossOrderRejectTransaction') is not None:
                parsed_body['trailingStopLossOrderRejectTransaction'] = \
                    transaction.TrailingStopLossOrderRejectTransaction.from_dict(
                        jbody['trailingStopLossOrderRejectTransaction']
                    )

            if jbody.get('lastTransactionID') is not None:
                parsed_body['lastTransactionID'] = \
                    jbody.get('lastTransactionID')

            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 401:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 404:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        if response.status is 405:
            if jbody.get('errorCode') is not None:
                parsed_body['errorCode'] = \
                    jbody.get('errorCode')

            if jbody.get('errorMessage') is not None:
                parsed_body['errorMessage'] = \
                    jbody.get('errorMessage')


        response.body = parsed_body

        return response

